#include <linux/module.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "./ti.h"

int tiopenlock_0;
int tiopenlock_1;
int tiopenlock_P0;
int tiopenlock_P1;
int tiopenlock_P2;
int tiopenlock_S0;
int tiopenlock_S1;
int tiopenlock_S2;
int tiopenlock_S3;

volatile char ti_0_buf[TIBUFSIZE];
volatile int ti_0_head, ti_0_tail;

volatile char ti_1_buf[TIBUFSIZE];
volatile int ti_1_head, ti_1_tail;

int parallel_port_table[3] = {0x3bc, 0x378, 0x278};
int serial_port_table[4] = {0x3f8, 0x2f8, 0x3e8, 0x2e8};

static int release_ti(struct inode * inode, struct file * file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  MOD_DEC_USE_COUNT;
  switch (minor) {
  case TI_MINOR_0: tiopenlock_0--; return 0;
  case TI_MINOR_1: tiopenlock_1--; return 0;
  case TI_MINOR_P0: tiopenlock_P0--; return 0;
  case TI_MINOR_P1: tiopenlock_P1--; return 0;
  case TI_MINOR_P2: tiopenlock_P2--; return 0;
  case TI_MINOR_S0: tiopenlock_S0--; return 0;
  case TI_MINOR_S1: tiopenlock_S1--; return 0;
  case TI_MINOR_S2: tiopenlock_S2--; return 0;
  case TI_MINOR_S3: tiopenlock_S3--; return 0;
  }
  return 0;
}

static int open_ti(struct inode * inode, struct file * file)
{
  unsigned int minor = MINOR(inode->i_rdev);
  MOD_INC_USE_COUNT;
  switch (minor) {
  case TI_MINOR_0:
    if (tiopenlock_0++)
      ti_0_tail = ti_0_head = 0;
    return 0;
  case TI_MINOR_1:
    if (tiopenlock_1++)
      ti_1_tail = ti_1_head = 0;  
    return 0;
  case TI_MINOR_P0: tiopenlock_P0++; return 0;
  case TI_MINOR_P1: tiopenlock_P1++; printk(KERN_CRIT "err0\n"); return 0;
  case TI_MINOR_P2: tiopenlock_P2++; return 0;
  case TI_MINOR_S0: tiopenlock_S0++; return 0;
  case TI_MINOR_S1: tiopenlock_S1++; return 0;
  case TI_MINOR_S2: tiopenlock_S2++; return 0;
  case TI_MINOR_S3: tiopenlock_S3++; return 0;
  }
  return 0;
}

static ssize_t write_ti(struct file * file, const char * buf, size_t count, loff_t * ppos)
{
  ssize_t n;
  unsigned int i;
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
  switch (minor) {
  case TI_MINOR_0:
    if(!tiopenlock_1)
      return 0;
    for(n=0;n<count;n++) {
      if((i = (ti_0_head+1) % TIBUFSIZE) == ti_0_tail) break;
      get_user(ti_0_buf[ti_0_head = i],buf+n);
    }
    return n;
  case TI_MINOR_1:
    if(!tiopenlock_0)
      return 0;
    for(n=0;n<count;n++) {
      if((i = (ti_1_head+1) % TIBUFSIZE) == ti_1_tail) break;
      get_user(ti_1_buf[ti_1_head = i],buf+n);
    }
    return n;
  case TI_MINOR_P0:
  case TI_MINOR_P1:
  case TI_MINOR_P2:
    i = parallel_port_table[minor-TI_MINOR_P0];
    for(n=0;n<count;n++) {
      char c;
      get_user(c,buf+n);
      if(put_ti_parallel(i,c)==-1) {
	init_ti_parallel(i);
	break;
      }
    }
    return n;
  case TI_MINOR_S0:
  case TI_MINOR_S1:
  case TI_MINOR_S2:
  case TI_MINOR_S3:
    i = serial_port_table[minor-TI_MINOR_S0];
    for(n=0;n<count;n++) {
      char c;
      get_user(c,buf+n);
      if(put_ti_serial(i, c)==-1) {
	init_ti_serial(i);
	break;
      }
    }
    return n;
  }
  return 0;
}

static ssize_t read_ti(struct file * file, char * buf, size_t count, loff_t * ppos)
{
  int c;
  ssize_t n;
  unsigned int i;
  unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev); 
  /* this is an awfull hack due to the changes between 2.0.x and 2.2.x kernels */
  
  if (ppos != &file->f_pos) return -ESPIPE; 
  /* completely useless: just to 
   * make sure everything is clean
   * ppos is the offset within the 
   * opened file to implement pread 
   * and pwrite: new functiosn within 2.2
   * kernels ... of no use here. but well: cleaner :)
   */

  switch (minor) {
  case TI_MINOR_0:
    if(!tiopenlock_1)
      return 0;
    for(n=0;n<count;n++) {
      if(ti_1_tail == ti_1_head) break;
      put_user(ti_1_buf[ti_1_tail = (ti_1_tail+1) % TIBUFSIZE],buf+n);
    }
    return n;
  case TI_MINOR_1:
    if(!tiopenlock_0)
      return 0;
    for(n=0;n<count;n++) {
      if(ti_0_tail == ti_0_head) break;
      put_user(ti_0_buf[ti_0_tail = (ti_0_tail+1) % TIBUFSIZE],buf+n);
    }
    return n;
  case TI_MINOR_P0:
  case TI_MINOR_P1:
  case TI_MINOR_P2:
    i = parallel_port_table[minor-TI_MINOR_P0];
    for(n=0;n<count;n++) {
      /*if(!probe_parallel(i)) break;*/
	/*well, idon't know why this is necessary but it is so... 
	 * when enabled, the thing cannot get the bits from the parport
	 */
      if((c=get_ti_parallel(i))==-1) {
	init_ti_parallel(i);
	break;
      }
      put_user(c,buf+n);
    }
    return n;
  case TI_MINOR_S0:
  case TI_MINOR_S1:
  case TI_MINOR_S2:
  case TI_MINOR_S3:
    i = serial_port_table[minor-TI_MINOR_S0];
    for(n=0;n<count;n++) {
      if(!probe_serial(i)) break;
      if((c=get_ti_serial(i))==-1) {
	init_ti_serial(i);
	break;
      }
      put_user(c,buf+n);
    }
    return n;
  }
  return 0;
}

/*******************/
/* Module routines */
/*******************/

struct file_operations ti_fops = {
  NULL,         /* lseek */
  read_ti,      
  write_ti,     
  NULL,         /* readdir */
  NULL,         /* poll */
  NULL,         /* ioctl */
  NULL,         /* mmap */
  open_ti,      
  NULL,         /* flush */
  release_ti,   
  NULL,         /* fsync */ 
  NULL,         /* fasync */
  NULL,         /* check_media_change */
  NULL,         /* revalidate */
  NULL,         /* lock */
};

static int init_module(void)
{
  int i;

  if(register_chrdev(TI_MAJOR,"ti",&ti_fops)) {
    printk("unable to get major %d for TI\n",TI_MAJOR);
    return 1;
  }

  for (i=0; i<3; i++)
    if (check_region(parallel_port_table[i], 3) >= 0)
      request_region(parallel_port_table[i], 3, "ti");

  for (i=0; i<4; i++)
    if (check_region(serial_port_table[i], 8) >= 0)
      request_region(serial_port_table[i], 8, "ti");

  printk("TI module loaded\n");

  tiopenlock_0=0;
  tiopenlock_1=0;
  tiopenlock_P0=0;
  tiopenlock_P1=0;
  tiopenlock_P2=0;
  tiopenlock_S0=0;
  tiopenlock_S1=0;
  tiopenlock_S2=0;
  tiopenlock_S3=0;

  return 0;
}

void cleanup_module(void)
{
  int i;

  unregister_chrdev(TI_MAJOR,"ti");

  for (i=0; i<3; i++)
    release_region(parallel_port_table[i], 3);

  for (i=0; i<4; i++)
    release_region(serial_port_table[i], 8);

  printk("TI module unloaded\n");
}

/**************************/
/* Parallel link routines */
/**************************/

int init_ti_parallel(int lpt_base)
{
  outb(3, lpt_base+0);
  return 0;
}

int put_ti_parallel(int lpt_base, unsigned char data)
{
  int bit,i,titime=0;
	
  for (bit=0; bit<8; bit++) {
    for(i=0; i<IO_DELAY; i++) inb(lpt_base+1);
    schedule();
    if (data&1) {
      outb(2, lpt_base+0);
      while (inb(lpt_base+1)&0x10) if(titime++>TIMAXTIME) return -1;
      outb(3, lpt_base+0);
      while (!(inb(lpt_base+1)&0x10)) if(titime++>TIMAXTIME) return -1;
    } else {
      outb(1, lpt_base+0);
      while (inb(lpt_base+1)&0x20) if(titime++>TIMAXTIME) return -1;
      outb(3, lpt_base+0);
      while (!(inb(lpt_base+1)&0x20)) if(titime++>TIMAXTIME) return -1;
    }
    data>>=1;
  }
  return 0;
}

int probe_parallel(int lpt_base)
{
  if((inb(lpt_base+1)&0x30)==0x30) return 0;
  return 1;
}

int get_ti_parallel(int lpt_base)
{
  int bit,i,titime=0;
  unsigned char data=0, v;

  for (bit=0; bit<8; bit++) {
    while ((v=inb(lpt_base+1)&0x30)==0x30) if(titime++>TIMAXTIME) return -1;

    if (v==0x10) {
      data=(data>>1)|0x80;
      outb(1, lpt_base+0);
      while (!(inb(lpt_base+1)&0x20)) if(titime++>TIMAXTIME) return -1;
      outb(3, lpt_base+0);
    } else {
      data=data>>1;
      outb(2, lpt_base+0);
      while (!(inb(lpt_base+1)&0x10)) if(titime++>TIMAXTIME) return -1;
      outb(3, lpt_base+0);
    }
    for(i=0; i<IO_DELAY; i++) inb(lpt_base+1);
    schedule();
  }
  return (int)data;
}

/************************/
/* Serial link routines */
/************************/

int init_ti_serial(int com_base)
{
  outb(3, com_base+4);
  return 0;
}

int put_ti_serial(int com_base, unsigned char data)
{
  int bit,i,titime=0;
	
  for (bit=0; bit<8; bit++) {
    for(i=0; i<IO_DELAY; i++) inb(com_base+6);
    schedule();
    if (data&1) {
      outb(2, com_base+4);
      while (inb(com_base+6)&0x10) if(titime++>TIMAXTIME) return -1;
      outb(3, com_base+4);
      while (!(inb(com_base+6)&0x10)) if(titime++>TIMAXTIME) return -1;
    } else {
      outb(1, com_base+4);
      while (inb(com_base+6)&0x20) if(titime++>TIMAXTIME) return -1;
      outb(3, com_base+4);
      while (!(inb(com_base+6)&0x20)) if(titime++>TIMAXTIME) return -1;
    }
    data>>=1;
  }
  return 0;
}

int probe_serial(int com_base)
{
  if((inb(com_base+6)&0x30)==0x30) return 0;
  return 1;
}

int get_ti_serial(int com_base)
{
  int bit,i,titime=0;
  unsigned char data=0, v;
  
  for (bit=0; bit<8; bit++) {
    while ((v=inb(com_base+6)&0x30)==0x30) if(titime++>TIMAXTIME) return -1;
    if (v==0x10) {
      data=(data>>1)|0x80;
      outb(1, com_base+4);
      while (!(inb(com_base+6)&0x20)) if(titime++>TIMAXTIME) return -1;
      outb(3, com_base+4);
    } else {
      data=data>>1;
      outb(2, com_base+4);
      while (!(inb(com_base+6)&0x10)) if(titime++>TIMAXTIME) return -1;
      outb(3, com_base+4);
    }
    for(i=0; i<IO_DELAY; i++) inb(com_base+6);
    schedule();
  }
  return (int)data;
}

